@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997, 1998, 1999 Kungliga Tekniska Hgskolan
+ * Copyright (c) 1997 - 2003 Kungliga Tekniska Hgskolan
  * (Royal Institute of Technology, Stockholm, Sweden). 
  * All rights reserved. 
  *
@@ -14,12 +14,7 @@
  *    notice, this list of conditions and the following disclaimer in the 
  *    documentation and/or other materials provided with the distribution. 
  *
- * 3. All advertising materials mentioning features or use of this software 
- *    must display the following acknowledgement: 
- *      This product includes software developed by Kungliga Tekniska 
- *      Hgskolan and its contributors. 
- *
- * 4. Neither the name of the Institute nor the names of its contributors 
+ * 3. Neither the name of the Institute nor the names of its contributors 
  *    may be used to endorse or promote products derived from this software 
  *    without specific prior written permission. 
  *
@@ -38,12 +33,12 @@
 
 #include "krb5_locl.h"
 
-RCSID("$Id: init_creds_pw.c,v 1.28 1999/01/30 18:28:46 assar Exp $");
+RCSID("$Id: init_creds_pw.c,v 1.55 2003/03/20 18:07:31 lha Exp $");
 
 static int
 get_config_time (krb5_context context,
-		 char *realm,
-		 char *name,
+		 const char *realm,
+		 const char *name,
 		 int def)
 {
     int ret;
@@ -64,24 +59,6 @@ get_config_time (krb5_context context,
     return def;
 }
 
-static krb5_boolean
-get_config_bool (krb5_context context,
-		 char *realm,
-		 char *name)
-{
-    return krb5_config_get_bool (context,
-				 NULL,
-				 "realms",
-				 realm,
-				 name,
-				 NULL)
-	|| krb5_config_get_bool (context,
-				 NULL,
-				 "libdefaults",
-				 name,
-				 NULL);
-}
-
 static krb5_error_code
 init_cred (krb5_context context,
 	   krb5_creds *cred,
@@ -93,7 +70,7 @@ init_cred (krb5_context context,
     krb5_error_code ret;
     krb5_realm *client_realm;
     int tmp;
-    int32_t now;
+    krb5_timestamp now;
 
     krb5_timeofday (context, &now);
 
@@ -116,22 +93,13 @@ init_cred (krb5_context context,
     if (options->flags & KRB5_GET_INIT_CREDS_OPT_TKT_LIFE)
 	tmp = options->tkt_life;
     else
-	tmp = get_config_time (context,
-			       *client_realm,
-			       "ticket_lifetime",
-			       10 * 60 * 60);
+	tmp = 10 * 60 * 60;
     cred->times.endtime = now + tmp;
 
-    tmp = 0;
-    if (options->flags & KRB5_GET_INIT_CREDS_OPT_RENEW_LIFE)
-	tmp = options->renew_life;
-    else
-	tmp = get_config_time (context,
-			       *client_realm,
-			       "renew_lifetime",
-			       0);
-    if (tmp)
-	cred->times.renew_till = now + tmp;
+    if ((options->flags & KRB5_GET_INIT_CREDS_OPT_RENEW_LIFE) &&
+	options->renew_life > 0) {
+	cred->times.renew_till = now + options->renew_life;
+    }
 
     if (in_tkt_service) {
 	krb5_realm server_realm;
@@ -140,7 +108,7 @@ init_cred (krb5_context context,
 	if (ret)
 	    goto out;
 	server_realm = strdup (*client_realm);
-	free (cred->server->realm);
+	free (*krb5_princ_realm(context, cred->server));
 	krb5_princ_set_realm (context, cred->server, &server_realm);
     } else {
 	ret = krb5_make_principal(context, &cred->server, 
@@ -156,6 +124,24 @@ out:
     return ret;
 }
 
+/*
+ * Print a message (str) to the user about the expiration in `lr'
+ */
+
+static void
+report_expiration (krb5_context context,
+		   krb5_prompter_fct prompter,
+		   krb5_data *data,
+		   const char *str,
+		   time_t time)
+{
+    char *p;
+	    
+    asprintf (&p, "%s%s", str, ctime(&time));
+    (*prompter) (context, data, NULL, p, 0, NULL);
+    free (p);
+}
+
 /*
  * Parse the last_req data and show it to the user if it's interesting
  */
@@ -169,8 +155,9 @@ print_expire (krb5_context context,
 {
     int i;
     LastReq *lr = &rep->enc_part.last_req;
-    int32_t sec;
+    krb5_timestamp sec;
     time_t t;
+    krb5_boolean reported = FALSE;
 
     krb5_timeofday (context, &sec);
 
@@ -180,26 +167,30 @@ print_expire (krb5_context context,
 			       7 * 24 * 60 * 60);
 
     for (i = 0; i < lr->len; ++i) {
-	if (lr->val[i].lr_type == 6
-	    && lr->val[i].lr_value <= t) {
-	    char *p;
-	    
-	    asprintf (&p, "Your password will expire at %s",
-		      ctime(&lr->val[i].lr_value));
-	    (*prompter) (context, data, p, 0, NULL);
-	    free (p);
-	    return;
+	if (lr->val[i].lr_value <= t) {
+	    switch (abs(lr->val[i].lr_type)) {
+	    case LR_PW_EXPTIME :
+		report_expiration(context, prompter, data,
+				  "Your password will expire at ",
+				  lr->val[i].lr_value);
+		reported = TRUE;
+		break;
+	    case LR_ACCT_EXPTIME :
+		report_expiration(context, prompter, data,
+				  "Your account will expire at ",
+				  lr->val[i].lr_value);
+		reported = TRUE;
+		break;
+	    }
 	}
     }
 
-    if (rep->enc_part.key_expiration
+    if (!reported
+	&& rep->enc_part.key_expiration
 	&& *rep->enc_part.key_expiration <= t) {
-	char *p;
-
-	asprintf (&p, "Your password/account will expire at %s",
-		  ctime(rep->enc_part.key_expiration));
-	(*prompter) (context, data, p, 0, NULL);
-	free (p);
+	report_expiration(context, prompter, data,
+			  "Your password/account will expire at ",
+			  *rep->enc_part.key_expiration);
     }
 }
 
@@ -218,6 +209,12 @@ get_init_creds_common(krb5_context context,
 {
     krb5_error_code ret;
     krb5_realm *client_realm;
+    krb5_get_init_creds_opt default_opt;
+
+    if (options == NULL) {
+	krb5_get_init_creds_opt_init (&default_opt);
+	options = &default_opt;
+    }
 
     ret = init_cred (context, cred, client, start_time,
 		     in_tkt_service, options);
@@ -230,18 +227,12 @@ get_init_creds_common(krb5_context context,
 
     if (options->flags & KRB5_GET_INIT_CREDS_OPT_FORWARDABLE)
 	flags->b.forwardable = options->forwardable;
-    else
-	flags->b.forwardable = get_config_bool (context,
-						*client_realm,
-						"forwardable");
 
     if (options->flags & KRB5_GET_INIT_CREDS_OPT_PROXIABLE)
 	flags->b.proxiable = options->proxiable;
-    else
-	flags->b.proxiable = get_config_bool (context,
-					      *client_realm,
-					      "proxiable");
 
+    if (start_time)
+	flags->b.postdated = 1;
     if (cred->times.renew_till)
 	flags->b.renewable = 1;
     if (options->flags & KRB5_GET_INIT_CREDS_OPT_ADDRESS_LIST)
@@ -249,8 +240,10 @@ get_init_creds_common(krb5_context context,
     if (options->flags & KRB5_GET_INIT_CREDS_OPT_ETYPE_LIST) {
 	*etypes = malloc((options->etype_list_length + 1)
 			* sizeof(krb5_enctype));
-	if (*etypes == NULL)
+	if (*etypes == NULL) {
+	    krb5_set_error_string(context, "malloc: out of memory");
 	    return ENOMEM;
+	}
 	memcpy (*etypes, options->etype_list,
 		options->etype_list_length * sizeof(krb5_enctype));
 	(*etypes)[options->etype_list_length] = ETYPE_NULL;
@@ -258,14 +251,18 @@ get_init_creds_common(krb5_context context,
     if (options->flags & KRB5_GET_INIT_CREDS_OPT_PREAUTH_LIST) {
 	*pre_auth_types = malloc((options->preauth_list_length + 1)
 				 * sizeof(krb5_preauthtype));
-	if (*pre_auth_types == NULL)
+	if (*pre_auth_types == NULL) {
+	    krb5_set_error_string(context, "malloc: out of memory");
 	    return ENOMEM;
+	}
 	memcpy (*pre_auth_types, options->preauth_list,
 		options->preauth_list_length * sizeof(krb5_preauthtype));
 	(*pre_auth_types)[options->preauth_list_length] = KRB5_PADATA_NONE;
     }
     if (options->flags & KRB5_GET_INIT_CREDS_OPT_SALT)
 	;			/* XXX */
+    if (options->flags & KRB5_GET_INIT_CREDS_OPT_ANONYMOUS)
+	flags->b.request_anonymous = options->anonymous;
     return 0;
 }
 
@@ -279,11 +276,11 @@ change_password (krb5_context context,
 		 void *data,
 		 krb5_get_init_creds_opt *old_options)
 {
-    krb5_prompt prompt;
+    krb5_prompt prompts[2];
     krb5_error_code ret;
     krb5_creds cpw_cred;
     char buf1[BUFSIZ], buf2[BUFSIZ];
-    krb5_data password_data;
+    krb5_data password_data[2];
     int result_code;
     krb5_data result_code_string;
     krb5_data result_string;
@@ -294,9 +291,12 @@ change_password (krb5_context context,
 
     krb5_get_init_creds_opt_init (&options);
     krb5_get_init_creds_opt_set_tkt_life (&options, 60);
-    krb5_get_init_creds_opt_set_preauth_list (&options,
-					      old_options->preauth_list,
-					      old_options->preauth_list_length);					      
+    krb5_get_init_creds_opt_set_forwardable (&options, FALSE);
+    krb5_get_init_creds_opt_set_proxiable (&options, FALSE);
+    if (old_options->flags & KRB5_GET_INIT_CREDS_OPT_PREAUTH_LIST)
+	krb5_get_init_creds_opt_set_preauth_list (&options,
+						  old_options->preauth_list,
+						  old_options->preauth_list_length);					      
 
     krb5_data_zero (&result_code_string);
     krb5_data_zero (&result_string);
@@ -314,30 +314,34 @@ change_password (krb5_context context,
 	goto out;
 
     for(;;) {
-	password_data.data   = buf1;
-	password_data.length = sizeof(buf1);
+	password_data[0].data   = buf1;
+	password_data[0].length = sizeof(buf1);
 
-	prompt.hidden = 1;
-	prompt.prompt = "New password";
-	prompt.reply  = &password_data;
+	prompts[0].hidden = 1;
+	prompts[0].prompt = "New password: ";
+	prompts[0].reply  = &password_data[0];
+	prompts[0].type   = KRB5_PROMPT_TYPE_NEW_PASSWORD;
 
-	ret = (*prompter) (context, data, "Changing password", 1, &prompt);
-	if (ret)
-	    goto out;
+	password_data[1].data   = buf2;
+	password_data[1].length = sizeof(buf2);
 
-	password_data.data   = buf2;
-	password_data.length = sizeof(buf2);
+	prompts[1].hidden = 1;
+	prompts[1].prompt = "Repeat new password: ";
+	prompts[1].reply  = &password_data[1];
+	prompts[1].type   = KRB5_PROMPT_TYPE_NEW_PASSWORD_AGAIN;
 
-	prompt.hidden = 1;
-	prompt.prompt = "Repeat new password";
-	prompt.reply  = &password_data;
-
-	ret = (*prompter) (context, data, "Changing password", 1, &prompt);
-	if (ret)
+	ret = (*prompter) (context, data, NULL, "Changing password",
+			   2, prompts);
+	if (ret) {
+	    memset (buf1, 0, sizeof(buf1));
+	    memset (buf2, 0, sizeof(buf2));
 	    goto out;
+	}
 
 	if (strcmp (buf1, buf2) == 0)
 	    break;
+	memset (buf1, 0, sizeof(buf1));
+	memset (buf2, 0, sizeof(buf2));
     }
     
     ret = krb5_change_password (context,
@@ -353,13 +357,15 @@ change_password (krb5_context context,
 	      (int)result_string.length,
 	      (char*)result_string.data);
 
-    ret = (*prompter) (context, data, p, 0, NULL);
+    ret = (*prompter) (context, data, NULL, p, 0, NULL);
     free (p);
     if (result_code == 0) {
-	strncpy (newpw, buf1, newpw_sz);
+	strlcpy (newpw, buf1, newpw_sz);
 	ret = 0;
-    } else
+    } else {
+	krb5_set_error_string (context, "failed changing password");
 	ret = ENOTTY;
+    }
 
 out:
     memset (buf1, 0, sizeof(buf1));
@@ -386,14 +392,12 @@ krb5_get_init_creds_password(krb5_context context,
     krb5_addresses *addrs = NULL;
     krb5_enctype *etypes = NULL;
     krb5_preauthtype *pre_auth_types = NULL;
-    krb5_preauthdata *preauth = NULL, preauth2;
     krb5_creds this_cred;
     krb5_kdc_rep kdc_reply;
     char buf[BUFSIZ];
     krb5_data password_data;
     int done;
 
-    memset(&kdc_reply, 0, sizeof(kdc_reply));
     ret = get_init_creds_common(context, creds, client, start_time,
 				in_tkt_service, options,
 				&addrs, &etypes, &this_cred, &pre_auth_types,
@@ -403,21 +407,24 @@ krb5_get_init_creds_password(krb5_context context,
 
     if (password == NULL) {
 	krb5_prompt prompt;
-	char *p;
+	char *p, *q;
 
 	krb5_unparse_name (context, this_cred.client, &p);
-	asprintf (&prompt.prompt, "%s's Password: ", p);
+	asprintf (&q, "%s's Password: ", p);
 	free (p);
+	prompt.prompt = q;
 	password_data.data   = buf;
 	password_data.length = sizeof(buf);
 	prompt.hidden = 1;
 	prompt.reply  = &password_data;
+	prompt.type   = KRB5_PROMPT_TYPE_PASSWORD;
 
-	ret = (*prompter) (context, data, NULL, 1, &prompt);
-	free (prompt.prompt);
+	ret = (*prompter) (context, data, NULL, NULL, 1, &prompt);
+	free (q);
 	if (ret) {
 	    memset (buf, 0, sizeof(buf));
-	    ret = KRB5KDC_ERR_NONE;
+	    ret = KRB5_LIBOS_PWDINTR;
+	    krb5_clear_error_string (context);
 	    goto out;
 	}
 	password = password_data.data;
@@ -425,12 +432,13 @@ krb5_get_init_creds_password(krb5_context context,
 
     done = 0;
     while(!done) {
+	memset(&kdc_reply, 0, sizeof(kdc_reply));
 	ret = krb5_get_in_cred (context,
 				flags.i,
 				addrs,
 				etypes,
 				pre_auth_types,
-				preauth,
+				NULL,
 				krb5_password_key_proc,
 				password,
 				NULL,
@@ -442,6 +450,17 @@ krb5_get_init_creds_password(krb5_context context,
 	    done = 1;
 	    break;
 	case KRB5KDC_ERR_KEY_EXPIRED :
+	    /* try to avoid recursion */
+
+	    if (prompter == NULL)
+		goto out;
+
+	    krb5_clear_error_string (context);
+
+	    if (in_tkt_service != NULL
+		&& strcmp (in_tkt_service, "kadmin/changepw") == 0)
+		goto out;
+
 	    ret = change_password (context,
 				   client,
 				   password,
@@ -454,56 +473,9 @@ krb5_get_init_creds_password(krb5_context context,
 		goto out;
 	    password = buf;
 	    break;
-	case KRB5KDC_ERR_PREAUTH_REQUIRED : {
-	    if(kdc_reply.error.e_data){
-		METHOD_DATA md;
-		int i;
-		decode_METHOD_DATA(kdc_reply.error.e_data->data, 
-				   kdc_reply.error.e_data->length, 
-				   &md, 
-				   NULL);
-		for(i = 0; i < md.len; i++){
-		    switch(md.val[i].padata_type){
-		    case pa_enc_timestamp:
-			if (pre_auth_types)
-			    free (pre_auth_types);
-			ALLOC(pre_auth_types, 2);
-			if (pre_auth_types == NULL) {
-			    free_METHOD_DATA(&md);
-			    goto out;
-			}
-			pre_auth_types[0] = KRB5_PADATA_ENC_TIMESTAMP;
-			pre_auth_types[1] = KRB5_PADATA_NONE;
-			break;
-		    case pa_etype_info:
-			preauth = &preauth2;
-			ALLOC_SEQ(preauth, 1);
-			preauth->val[0].type = KRB5_PADATA_ENC_TIMESTAMP;
-			krb5_decode_ETYPE_INFO(context,
-					       md.val[i].padata_value.data, 
-					       md.val[i].padata_value.length,
-					       &preauth->val[0].info,
-					       NULL);
-			break;
-		    }
-		}
-		free_METHOD_DATA(&md);
-	    }else{
-		if (pre_auth_types)
-		    free (pre_auth_types);
-		pre_auth_types = malloc(2 * sizeof(*pre_auth_types));
-		if (pre_auth_types == NULL)
-		    goto out;
-		pre_auth_types[0] = KRB5_PADATA_ENC_TIMESTAMP;
-		pre_auth_types[1] = KRB5_PADATA_NONE;
-	    }
-	    krb5_free_kdc_rep (context, &kdc_reply);
-	    break;
-	}
 	default:
 	    goto out;
 	}
-	memset(&kdc_reply, 0, sizeof(kdc_reply));
     }
 
     if (prompter)
@@ -514,13 +486,10 @@ krb5_get_init_creds_password(krb5_context context,
 		      data);
 out:
     memset (buf, 0, sizeof(buf));
-    krb5_free_kdc_rep (context, &kdc_reply);
+    if (ret == 0)
+	krb5_free_kdc_rep (context, &kdc_reply);
 
     free (pre_auth_types);
-    if(preauth) {
-	free_ETYPE_INFO(&preauth->val[0].info);
-	free(preauth->val);
-    }
     free (etypes);
     if (ret == 0 && creds)
 	*creds = this_cred;
@@ -565,6 +534,7 @@ krb5_get_init_creds_keytab(krb5_context context,
 
     a = malloc (sizeof(*a));
     if (a == NULL) {
+	krb5_set_error_string(context, "malloc: out of memory");
 	ret = ENOMEM;
 	goto out;
     }
@@ -583,6 +553,8 @@ krb5_get_init_creds_keytab(krb5_context context,
 			    NULL,
 			    &this_cred,
 			    NULL);
+    free (a);
+
     if (ret)
 	goto out;
     free (pre_auth_types);
